﻿using System;
using System.IO;
using NAudio.Wave;

namespace WpfAudioPlayer.Models.NAudio
{
	/// <summary>
	/// NAudio を利用して音楽再生を行います。
	/// </summary>
	sealed class AudioPlayer : IAudioPlayer
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="fileName">音楽ファイルへのパス。</param>
		/// <exception cref="FileNotFoundException">音楽ファイルが存在しない。</exception>
		/// <exception cref="Exception">ストリームの生成に失敗した。</exception>
		public AudioPlayer( string fileName )
		{
			if( !File.Exists( fileName ) ) { throw new FileNotFoundException( fileName ); }

			try
			{
				this.InitializeStream( fileName );
			}
			catch( Exception exp )
			{
				this.Dispose();
				throw exp;
			}
		}

		/// <summary>
		/// 音楽ファイルの音声データへのストリームを生成します。
		/// </summary>
		/// <param name="fileName">音楽ファイルへのパス。</param>
		/// <exception cref="InvalidOperationException">ストリームの生成に失敗した。</exception>
		private void InitializeStream( string fileName )
		{
			WaveChannel32 stream;
			if( fileName.EndsWith( ".wav" ) )
			{
				WaveStream reader = new WaveFileReader( fileName );
				if( reader.WaveFormat.Encoding != WaveFormatEncoding.Pcm )
				{
					reader = WaveFormatConversionStream.CreatePcmStream( reader );
					reader = new BlockAlignReductionStream( reader );
				}
				
				if( reader.WaveFormat.BitsPerSample != 16 )
				{
					var format = new WaveFormat( reader.WaveFormat.SampleRate, 16, reader.WaveFormat.Channels );
					reader = new WaveFormatConversionStream( format, reader );
				}
				
				stream = new WaveChannel32( reader );
			}
			else if( fileName.EndsWith( ".mp3" ) )
			{
				var reader             = new Mp3FileReader( fileName );
				var pcmStream          = WaveFormatConversionStream.CreatePcmStream( reader );
				var blockAlignedStream = new BlockAlignReductionStream( pcmStream );

				stream = new WaveChannel32( blockAlignedStream );
			}
			else
			{
				throw new InvalidOperationException( "Unsupported extension" );
			}

			this._volumeStream = stream;
			this._audioStream  = new MeteringStream( stream, stream.WaveFormat.SampleRate / 10 );

			this._waveOut = new WaveOut() { DesiredLatency = 200 };
			this._waveOut.Init( this._audioStream );
		}

		#region フィールド

		/// <summary>
		/// 音楽ファイルの音声データへのストリーム。
		/// </summary>
		private WaveStream _audioStream;

		/// <summary>
		/// 音量調整を行う為のストリーム。
		/// </summary>
		private WaveChannel32 _volumeStream;

		/// <summary>
		/// 音声の再生を行う為のオブジェクト。
		/// </summary>
		private IWavePlayer _waveOut;

		#endregion

		#region IAudioPlayer メンバ

		/// <summary>
		/// 再生を一時停止します。
		/// </summary>
		public void Pause()
		{
			this._waveOut.Pause();
		}

		/// <summary>
		/// 再生を開始します。
		/// </summary>
		public void Play()
		{
			switch( this._waveOut.PlaybackState )
			{
			case PlaybackState.Playing:
				break;
			
			case PlaybackState.Paused:
			case PlaybackState.Stopped:
				this._waveOut.Play();
				break;
			}
		}

		/// <summary>
		/// 再生を停止します。
		/// </summary>
		public void Stop()
		{
			this._waveOut.Stop();
			this._audioStream.Position = 0;
		}

		/// <summary>
		/// 再生位置の変更を行える事を示す値を取得します。
		/// </summary>
		public bool CanSeek
		{
			get
			{
				return this._audioStream.CanSeek;
			}
		}

		/// <summary>
		/// 再生位置を時間単位で取得または設定します。
		/// </summary>
		public TimeSpan CurrentTime
		{
			get
			{
				return this._audioStream.CurrentTime;
			}
			set
			{
				this._audioStream.CurrentTime = value;
			}
		}

		/// <summary>
		/// 演奏時間を取得します。
		/// </summary>
		public TimeSpan Duration
		{
			get
			{
				return this._audioStream.TotalTime;
			}
		}

		/// <summary>
		/// 音楽再生の状態を取得します。
		/// </summary>
		public PlayState PlayState
		{
			get
			{
				switch( this._waveOut.PlaybackState )
				{
				case PlaybackState.Paused:	return PlayState.Paused;
				case PlaybackState.Playing:	return PlayState.Playing;
				default:					return PlayState.Stopped;
				}
			}
		}

		/// <summary>
		/// 音量を取得または設定します。
		/// </summary>
		public float Volume
		{
			get
			{
				return this._volumeStream.Volume;
			}
			set
			{
				this._volumeStream.Volume = value;
			}
		}

		#endregion

		#region IDisposable メンバ

		/// <summary>
		/// リソースの解放を行います。
		/// </summary>
		public void Dispose()
		{
			if( this._waveOut != null )
			{
				this._waveOut.Stop();
			}

			if( this._audioStream != null )
			{
				this._volumeStream.Close();
				this._volumeStream = null;

				this._audioStream.Close();
				this._audioStream = null;
			}

			if( this._waveOut != null )
			{
				this._waveOut.Dispose();
				this._waveOut = null;
			}
		}

		#endregion
	}
}
