﻿using System;
using System.Globalization;
using System.Windows.Input;
using System.Windows.Threading;
using WpfAudioPlayer.Commands;
using WpfAudioPlayer.Models;

namespace WpfAudioPlayer.ViewModels
{
	/// <summary>
	/// プレイヤーの View と Model を仲介するクラスです。
	/// </summary>
	sealed class PlayerViewModel : ViewModelBase, IDisposable
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		public PlayerViewModel()
		{
			this._timer          = new DispatcherTimer();
			this._timer.Tick    += new EventHandler( OnTick );
			this._timer.Interval = TimeSpan.FromSeconds( 1.0 );

			this._volume = 1.0f;

			this.InitializeCommand();
		}

		/// <summary>
		/// 音楽再生を行う為のオブジェクトを生成します。
		/// </summary>
		/// <param name="fileName">音楽ファイルへのパス。</param>
		/// <returns>生成したオブジェクト。</returns>
		private static IAudioPlayer CreatePlayer( string fileName )
		{
			return new WpfAudioPlayer.Models.NAudio.AudioPlayer( fileName );
		}
	
		/// <summary>
		/// 次に再生対象とする曲を取得する。
		/// </summary>
		/// <param name="isNext">次の曲を要求する場合は true、前の曲ならば false。</param>
		/// <returns>次に再生対象とする曲。</returns>
		private SongViewModel GetNextSong( bool isNext )
		{
			if( this.RequestNextSong == null ) { return null; }

			return this.RequestNextSong( this, new RequestNextSongEventArgs( isNext, this.CurrentSong ) );
		}

		/// <summary>
		/// コマンドを初期化します。
		/// </summary>
		private void InitializeCommand()
		{
			// 再生
			this.PlayCommand = new DelegateCommand(
				() =>
				{
					this.PlayPause();
				},
				() =>
				{
					return ( this._audioPlayer != null );
				}
				);

			// 停止
			this.StopCommand = new DelegateCommand(
				() =>
				{
					this.Stop();
				},
				() =>
				{
					return ( this._audioPlayer != null && this._audioPlayer.PlayState != PlayState.Stopped );
				}
				);

			// 前へ
			this.BackCommand = new DelegateCommand(
				() =>
				{
					this.MoveSong( false );
				},
				() =>
				{
					return ( this.CurrentSong != null );
				}
				);

			// 次へ
			this.NextCommand = new DelegateCommand(
				() =>
				{
					this.MoveSong( true );
				},
				() =>
				{
					return ( this.CurrentSong != null );
				}
				);
		}

		/// <summary>
		/// 次に再生対象とする曲を選択します。
		/// </summary>
		/// <param name="isNext">次の曲を要求する場合は true、前の曲ならば false。</param>
		/// <returns>次に再生対象とする曲。</returns>
		private void MoveSong( bool isNext )
		{
			var song = this.GetNextSong( isNext );
			if( song != null )
			{
				if( this.PlayState == PlayState.Stopped )
				{
					this.CurrentSong = song;
				}
				else
				{
					this.Stop();
					this.CurrentSong = song;
					this.PlayPause();
				}
			}
		}

		/// <summary>
		/// 音声の再生中に呼び出されるタイマー イベント。
		/// </summary>
		/// <param name="sender">イベント送信元。</param>
		/// <param name="e">イベントデータ。</param>
		private void OnTick( object sender, EventArgs e )
		{
			if( this._audioPlayer == null ) { return; }

			if( this._audioPlayer.CurrentTime >= this._audioPlayer.Duration )
			{
				this.Stop();
			}
			else
			{
				this._currentTime = PlayerViewModel.TimeToText( this._audioPlayer.CurrentTime );
				this.NotifyPropertyChanged( "Position", "CurrentTime" );
			}
		}

		/// <summary>
		/// コンテンツの再生・一時停止を行います。
		/// </summary>
		private void PlayPause()
		{
			if( this._audioPlayer.PlayState == PlayState.Playing )
			{
				this._audioPlayer.Pause();
				this.IsPlaying = false;
				this.NotifyPropertyChanged( "IsPlaying" );
			}
			else
			{
				this._audioPlayer.Play();
				this._timer.Start();
				this.IsPlaying = true;
				this.NotifyPropertyChanged( "CanSeek", "IsPlaying" );
			}
		}

		/// <summary>
		/// コンテンツの再生を停止します。
		/// </summary>
		private void Stop()
		{
			if( this._audioPlayer != null )
			{
				this._timer.Stop();
				this._audioPlayer.Stop();
			}

			this._currentTime = PlayerViewModel.EmptyTimeText;
			this.IsPlaying = false;
			this.NotifyPropertyChanged( "CurrentTime", "CanSeek", "Position", "IsPlaying" );
		}

		/// <summary>
		/// 時間を文字列に変換します。
		/// </summary>
		/// <param name="time">時間。</param>
		/// <returns>変換された文字列。</returns>
		private static string TimeToText( TimeSpan time )
		{
			return String.Format( CultureInfo.InvariantCulture, "{0:d2}:{1:d2}:{2:d2}", time.Hours, time.Minutes, time.Seconds );
		}

		#region イベント

		/// <summary>
		/// 次に再生対象とする曲を要求するイベントハンドラ。
		/// </summary>
		/// <param name="sender">イベント送信元。</param>
		/// <param name="e">イベントデータ。</param>
		/// <returns>次に再生対象とする曲。</returns>
		public delegate SongViewModel RequestNextSongEventHandler( object sender, RequestNextSongEventArgs e );

		/// <summary>
		/// 次に再生すべき曲を要求するイベント。
		/// </summary>
		public event RequestNextSongEventHandler RequestNextSong;

		#endregion

		#region プロパティ

		/// <summary>
		/// コンテンツの再生位置が移動可能である事を示す値を取得します。
		/// </summary>
		public bool CanSeek
		{
			get
			{
				return ( this._audioPlayer == null ? false : ( this._audioPlayer.PlayState != PlayState.Stopped && this._audioPlayer.CanSeek ) );
			}
		}

		/// <summary>
		/// 現在選択されている曲情報を取得または設定します。
		/// </summary>
		public SongViewModel CurrentSong
		{
			get
			{
				return this._currentSong;
			}
			set
			{
				if( this.PlayState != PlayState.Stopped ) { return; }

				this._currentSong = value;
				if( this._currentSong == null ) { return; }

				if( this._audioPlayer != null )
				{
					this._audioPlayer.Dispose();
					this._audioPlayer = null;
				}

				this._audioPlayer = PlayerViewModel.CreatePlayer( value.FilePath );
				this._duration = PlayerViewModel.TimeToText( this._audioPlayer.Duration );

				this.Title = this._currentSong.Title;
				this.NotifyPropertyChanged( "Title", "Duration", "Length" );
			}
		}

		/// <summary>
		/// コンテンツの現在の再生時間を表す文字列を取得します。
		/// </summary>
		public string CurrentTime
		{
			get
			{
				return this._currentTime;
			}
		}

		/// <summary>
		/// コンテンツの総再生時間を表す文字列を取得します。
		/// </summary>
		public string Duration
		{
			get
			{
				return this._duration;
			}
		}

		/// <summary>
		/// 再生状態である事を示す値を取得または設定します。
		/// </summary>
		public bool IsPlaying { get; set; }

		/// <summary>
		/// コンテンツの総再生時間 ( 秒単位 ) を取得または設定します。
		/// </summary>
		public double Length
		{
			get
			{
				return ( this._audioPlayer == null ? double.MaxValue : this._audioPlayer.Duration.TotalSeconds );
			}
		}

		/// <summary>
		/// 音楽再生の状態を取得します。
		/// </summary>
		public PlayState PlayState
		{
			get
			{
				return ( this._audioPlayer == null ? PlayState.Stopped : this._audioPlayer.PlayState );
			}
		}

		/// <summary>
		/// コンテンツの再生位置 ( 秒単位 ) を取得または設定します。
		/// </summary>
		public double Position
		{
			get
			{
				return ( this._audioPlayer == null ? 0.0 : this._audioPlayer.CurrentTime.TotalSeconds );
			}
			set
			{
				if( this._audioPlayer != null )
				{
					this._audioPlayer.CurrentTime = TimeSpan.FromSeconds( value );
					this._currentTime = PlayerViewModel.TimeToText( this._audioPlayer.CurrentTime );
					this.NotifyPropertyChanged( "Position", "CurrentTime" );
				}
			}
		}

		/// <summary>
		/// コンテンツの題名を取得します。
		/// </summary>
		public string Title { get; private set; }

		/// <summary>
		/// コンテンツ再生時の音量を取得または設定します。
		/// </summary>
		public float Volume
		{
			get
			{
				return this._volume;
			}
			set
			{
				this._volume = value;
				if( this._audioPlayer != null )
				{
					this._audioPlayer.Volume = value;
				}
			}
		}

		#endregion

		#region コマンド

		/// <summary>
		/// 前のコンテンツを選択します。コンテンツが再生中の場合は、新たに選択されたコンテンツの再生に切り替えます。
		/// </summary>
		public ICommand BackCommand { get; private set; }

		/// <summary>
		/// 次のコンテンツを選択します。コンテンツが再生中の場合は、新たに選択されたコンテンツの再生に切り替えます。
		/// </summary>
		public ICommand NextCommand { get; private set; }

		/// <summary>
		/// コンテンツの再生を開始します。
		/// </summary>
		public ICommand PlayCommand { get; private set; }

		/// <summary>
		/// コンテンツの再生を停止します。
		/// </summary>
		public ICommand StopCommand { get; private set; }

		#endregion

		#region フィールド

		/// <summary>
		/// 無効な時間情報を表す文字列。
		/// </summary>
		private const string EmptyTimeText = "00:00:00";

		/// <summary>
		/// 現在選択されている曲情報。
		/// </summary>
		private SongViewModel _currentSong;

		/// <summary>
		/// コンテンツの現在の再生時間を表す文字列。
		/// </summary>
		private string _currentTime = PlayerViewModel.EmptyTimeText;

		/// <summary>
		/// コンテンツの現在の再生時間を表す文字列。
		/// </summary>
		private string _duration = PlayerViewModel.EmptyTimeText;

		/// <summary>
		/// 再生状態の定期監視を行う為のタイマー。
		/// </summary>
		private DispatcherTimer _timer;

		/// <summary>
		/// 音声再生を行う為のオブジェクト。
		/// </summary>
		private IAudioPlayer _audioPlayer;

		/// <summary>
		/// コンテンツ再生時の音量。
		/// </summary>
		private float _volume;

		#endregion

		#region IDisposable メンバ

		/// <summary>
		/// リソースの解放を行います。
		/// </summary>
		public void Dispose()
		{
			this.Stop();

			if( this._audioPlayer != null )
			{
				this._audioPlayer.Dispose();
				this._audioPlayer = null;
			}
		}

		#endregion
	}
	
	/// <summary>
	/// 次に再生対象とする曲を要求するイベントのデータ。
	/// </summary>
	class RequestNextSongEventArgs : EventArgs
	{
		/// <summary>
		/// インスタンスを初期化する。
		/// </summary>
		/// <param name="isNext">次の曲を要求する場合は true、前の曲ならば false。</param>
		/// <param name="isNext">現在、再生対象となっている曲。</param>
		public RequestNextSongEventArgs( bool isNext, SongViewModel song )
		{
			this.IsNext = isNext;
			this.Song = song;
		}

		/// <summary>
		/// 次の曲を要求する事を示す値を取得する。
		/// </summary>
		public bool IsNext { get; private set; }

		/// <summary>
		/// 現在、再生対象となっている曲を取得する。
		/// </summary>
		public SongViewModel Song { get; private set; }
	}
}
