﻿using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using WpfAudioPlayer.Models;

namespace WpfAudioPlayer.ViewModels
{
		/// <summary>
	/// 曲の View と Model を仲介するクラスです。
	/// </summary>
	class SongViewModel : ViewModelBase
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="song">曲情報。</param>
		public SongViewModel( Song song )
		{
			this._song = song;
		}

		/// <summary>
		/// 曲名を取得します。
		/// </summary>
		public string Title
		{
			get
			{
				return this._song.Title;
			}
		}

		/// <summary>
		/// 音楽ファイルへのパスを取得します。
		/// </summary>
		public string FilePath
		{
			get
			{
				return this._song.FilePath;
			}
		}

		/// <summary>
		/// 曲情報。
		/// </summary>
		private Song _song;
	}

	/// <summary>
	/// 曲一覧の View と Model を仲介するクラスです。
	/// </summary>
	class SongListViewModel : ViewModelBase
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		public SongListViewModel()
		{
			this.Songs = new ObservableCollection< SongViewModel >();
		}

		/// <summary>
		/// 音楽ファイルを追加します。
		/// </summary>
		/// <param name="fileNames">音楽ファイルのパス コレクション。</param>
		public void Add( string[] fileNames )
		{
			var worker = new BackgroundWorker() { WorkerReportsProgress = true };

			worker.DoWork += ( sender, e ) =>
			{
				foreach( var fileName in fileNames )
				{
					worker.ReportProgress( 0, Song.FromFile( fileName ) );
				}
			};

			worker.ProgressChanged += ( sender, e ) =>
			{
				var song = e.UserState as Song;
				if( song != null )
				{
					this.Songs.Add( new SongViewModel( song ) );
				}
			};

			worker.RunWorkerAsync();
		}

		/// <summary>
		/// 次の曲を取得します。
		/// </summary>
		/// <param name="isNext">次の曲を要求する場合は true、前の曲ならば false。</param>
		/// <param name="song">基準となる曲。</param>
		/// <returns>次の曲を表す SongViewModel インスタンス。次の曲が見つからない場合は null 参照。</returns>
		public SongViewModel GetNextSong( bool isNext, SongViewModel song )
		{
			if( song == null ) { return null; }

			int index = this.Songs.IndexOf( song ) + ( isNext ? 1 : -1 );
			if( index < 0 || this.Songs.Count == index ) { return null; }

			return this.Songs[ index ];	
		}

		/// <summary>
		/// 選択状態の曲情報を取得または設定します。
		/// </summary>
		public SongViewModel SelectedSong
		{
			get
			{
				return this._selectedSong;
			}
			set
			{
				this._selectedSong = value;
				if( this.SongSelecttionChanged != null )
				{
					this.SongSelecttionChanged( this, new SongSlecttionChangedEventArgs( value ) );
					this.NotifyPropertyChanged( "SelectedSong" );
				}
			}
		}

		/// <summary>
		/// 曲情報の選択状態が変更された時のイベント．
		/// </summary>
		public event EventHandler< SongSlecttionChangedEventArgs > SongSelecttionChanged;

		/// <summary>
		/// 曲情報のコレクションを取得します。
		/// </summary>
		public ObservableCollection< SongViewModel > Songs { get; private set; }

		/// <summary>
		/// 選択されている曲情報。
		/// </summary>
		private SongViewModel _selectedSong;
	}

	/// <summary>
	/// 曲情報が選択された時のイベント データを表します。
	/// </summary>
	class SongSlecttionChangedEventArgs : EventArgs
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="song">選択された曲情報。</param>
		public SongSlecttionChangedEventArgs( SongViewModel song )
		{
			this.Song = song;
		}

		/// <summary>
		/// 選択された曲情報を取得します。
		/// </summary>
		public SongViewModel Song { get; private set; }
	}

}
